@@ -42,6 +42,8 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 #define REQUEUE_MAX_SEC   (60 * 60)
 #define REQUEUE_SCALE     2
 
+#define GEOIP_URL         "geoip-url"
+
 typedef enum _MarkupField MarkupField;
 enum _MarkupField {
 	/* Address */
@@ -73,6 +75,7 @@ struct _UbuntuGeoipProviderPrivate
 {
 	SoupSession * soup;
 	NMClient * client;
+	GSettings * settings;
 
 	/* Markup thingies */
 	GMarkupParseContext * context;
@@ -91,7 +94,7 @@ struct _UbuntuGeoipProviderPrivate
 	gchar * timezone;
 
 	/* State variables */
-	gboolean connected;
+	NMState network_state;
 	gboolean gotdata;
 
 	/* Requeue variables */
@@ -149,6 +152,9 @@ static void markup_text                      (GMarkupParseContext * context,
                                               gsize text_len,
                                               gpointer user_data,
                                               GError ** error);
+static void url_changed                      (GSettings * settings,
+                                              char * key,
+                                              gpointer user_data);
 
 static GMarkupParser parser_struct = {
 	start_element: markup_start,
@@ -182,6 +188,22 @@ ubuntu_geoip_provider_class_init (UbuntuGeoipProviderClass *klass)
 	return;
 }
 
+/* Check to see if a schema exists */
+static gboolean
+settings_schema_exists (const gchar * schema)
+{
+	const gchar * const * schemas = g_settings_list_schemas();
+	int i;
+
+	for (i = 0; schemas[i] != NULL; i++) {
+		if (g_strcmp0(schemas[i], schema) == 0) {
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 /* Make everything consistent */
 static void
 ubuntu_geoip_provider_init (UbuntuGeoipProvider *self)
@@ -191,10 +213,11 @@ ubuntu_geoip_provider_init (UbuntuGeoipProvider *self)
 	/* Init */
 	self->priv->soup = NULL;
 	self->priv->client = NULL;
+	self->priv->settings = NULL;
 
 	self->priv->latitude = 0.0;
 	self->priv->longitude = 0.0;
-	self->priv->connected = FALSE;
+	self->priv->network_state = NM_STATE_UNKNOWN;
 	self->priv->gotdata = FALSE;
 
 	self->priv->country_code = NULL;
@@ -214,6 +237,12 @@ ubuntu_geoip_provider_init (UbuntuGeoipProvider *self)
 	/* Build Parser */
 	self->priv->context = g_markup_parse_context_new(&parser_struct, 0, &self->priv->markup_data, NULL);
 
+	/* Set up settings */
+	if (settings_schema_exists("com.ubuntu.geoip")) {
+		self->priv->settings = g_settings_new("com.ubuntu.geoip");
+		g_signal_connect(G_OBJECT(self->priv->settings), "changed::" GEOIP_URL, G_CALLBACK(url_changed), self);
+	}
+
 	/* Allocate */
 	self->priv->soup = soup_session_async_new();
 	self->priv->client = nm_client_new();
@@ -257,6 +286,11 @@ ubuntu_geoip_provider_dispose (GObject *object)
 		provider->priv->context = NULL;
 	}
 
+	if (provider->priv->settings != NULL) {
+		g_object_unref(provider->priv->settings);
+		provider->priv->settings = NULL;
+	}
+
 	G_OBJECT_CLASS (ubuntu_geoip_provider_parent_class)->dispose (object);
 	return;
 }
@@ -302,6 +336,25 @@ ubuntu_geoip_provider_finalize (GObject *object)
 	return;
 }
 
+/* Something could change our URL settings, if so we need to respond
+   and start getting the new data */
+static void
+url_changed (GSettings * settings, char * key, gpointer user_data)
+{
+	UbuntuGeoipProvider *provider = UBUNTU_GEOIP_PROVIDER(user_data);
+	g_return_if_fail(provider != NULL);
+
+	/* If we've got a network connection we care, otherwise we don't
+	   have a reason to use the new URL */
+	if (provider->priv->network_state == NM_STATE_CONNECTED_GLOBAL) {
+		provider->priv->gotdata = FALSE;
+		gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_ACQUIRING);
+		send_message(provider);
+	}
+
+	return;
+}
+
 /* Respond to a request to shutdown from the GeoClue provider
    interface. */
 static void
@@ -319,13 +372,17 @@ get_status (GcIfaceGeoclue *geoclue, GeoclueStatus * status, GError ** error)
 	UbuntuGeoipProvider *provider = UBUNTU_GEOIP_PROVIDER(geoclue);
 	g_return_val_if_fail(provider != NULL, FALSE);
 
-	if (provider->priv->connected == FALSE) {
+	if (provider->priv->network_state != NM_STATE_CONNECTED_GLOBAL &&
+		provider->priv->network_state != NM_STATE_CONNECTING) {
 		*status = GEOCLUE_STATUS_UNAVAILABLE;
+	} else if (provider->priv->network_state == NM_STATE_CONNECTING) {
+		*status = GEOCLUE_STATUS_ACQUIRING;
 	} else if (provider->priv->gotdata == FALSE) {
 		*status = GEOCLUE_STATUS_ACQUIRING;
 	} else {
 		*status = GEOCLUE_STATUS_AVAILABLE;
 	}
+
 	return TRUE;
 }
 
@@ -442,6 +499,10 @@ message_finished (SoupMessage * message, gpointer user_data)
 
 	if (g_markup_parse_context_parse(provider->priv->context, message->response_body->data, message->response_body->length, NULL) == FALSE) {
 		g_warning("Unable to parse the data from the geoip provider");
+
+		g_clear_object(&provider->priv->context);
+		provider->priv->context = g_markup_parse_context_new(&parser_struct, 0, &provider->priv->markup_data, NULL);
+
 		queue_message(provider);
 		return;
 	}
@@ -454,8 +515,9 @@ message_finished (SoupMessage * message, gpointer user_data)
 
 	provider->priv->requeue_time = REQUEUE_MIN_SEC;
 
+	gboolean emit = FALSE;
 	if (provider->priv->gotdata == FALSE) {
-		gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_AVAILABLE);
+		emit = TRUE;
 	}
 	provider->priv->gotdata = TRUE;
 
@@ -491,6 +553,11 @@ message_finished (SoupMessage * message, gpointer user_data)
 		geoclue_accuracy_free(accuracy);
 	}
 
+	if (emit) {
+		g_debug("Emitting Available");
+		gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_AVAILABLE);
+	}
+
 	return;
 }
 
@@ -505,6 +572,22 @@ message_complete (SoupSession * session, SoupMessage * message, gpointer user_da
 	return;
 }
 
+/* Handle events that can be errors */
+static void
+message_network_event (SoupMessage * message, GSocketClientEvent event, GIOStream * connection, gpointer user_data)
+{
+	UbuntuGeoipProvider *provider = UBUNTU_GEOIP_PROVIDER(user_data);
+	g_return_if_fail(provider != NULL);
+
+	if (event == G_SOCKET_CLIENT_COMPLETE && connection == NULL) {
+		/* This is the error condition */
+		g_warning("Network connection error");
+		queue_message(provider);
+	}
+
+	return;
+}
+
 /* Sends the message to the GeoIP service and sets up some
    callbacks for fun */
 static void
@@ -512,14 +595,30 @@ send_message (UbuntuGeoipProvider *provider)
 {
 	g_return_if_fail(provider->priv->soup != NULL);
 
+	/* Insure that we reset our timeout as we're fulfilling it */
 	if (provider->priv->requeue_source != 0) {
 		g_source_remove(provider->priv->requeue_source);
 		provider->priv->requeue_source = 0;
 	}
 
-	SoupMessage * message = soup_message_new("get", "http://geoip.ubuntu.com/lookup");
-	g_signal_connect(G_OBJECT(message), "finished", G_CALLBACK(message_finished), provider);
+	/* Grab the URL from the settings if we have it */
+	gchar * url = NULL;
+	if (provider->priv->settings != NULL) {
+		url = g_settings_get_string(provider->priv->settings, GEOIP_URL);
+	} else {
+		url = g_strdup("http://geoip.ubuntu.com/lookup");
+	}
+	g_debug("Using GeoIP URL: %s", url);
 
+	/* Build the message to send */
+	SoupMessage * message = soup_message_new("GET", url);
+	g_signal_connect(G_OBJECT(message), "finished", G_CALLBACK(message_finished), provider);
+	g_signal_connect(G_OBJECT(message), "network-event", G_CALLBACK(message_network_event), provider);
+
+	g_free(url);
+	url = NULL;
+
+	/* Ask SOUP to hit the interwebs for us */
 	soup_session_queue_message(provider->priv->soup, message,
 	                           message_complete, provider);
 
@@ -547,6 +646,7 @@ queue_message (UbuntuGeoipProvider * provider)
 		return;
 	}
 
+	g_debug("Queuing up new message");
 	provider->priv->requeue_time *= REQUEUE_SCALE;
 	if (provider->priv->requeue_time > REQUEUE_MAX_SEC) {
 		provider->priv->requeue_time = REQUEUE_MAX_SEC;
@@ -557,6 +657,94 @@ queue_message (UbuntuGeoipProvider * provider)
 	return;
 }
 
+/* Change the position to be zero and emit it if it changed */
+static void
+clear_position (UbuntuGeoipProvider * provider)
+{
+	if (provider->priv->latitude == 0.0 && provider->priv->longitude == 0.0) {
+		/* We didn't have any data */
+		return;
+	}
+
+	provider->priv->latitude = 0.0;
+	provider->priv->longitude = 0.0;
+
+	g_debug("Position Cleared");
+
+	GeoclueAccuracy * accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_NONE, 0.0, 0.0);
+	gc_iface_position_emit_position_changed(GC_IFACE_POSITION(provider), 
+	                                        time(NULL),
+	                                        GEOCLUE_POSITION_FIELDS_NONE,
+	                                        provider->priv->latitude,
+	                                        provider->priv->longitude,
+	                                        0, /* Altitude */
+	                                        accuracy);
+
+	geoclue_accuracy_free(accuracy);
+
+	return;
+}
+
+/* Change the address to be clear and emit it if it changed */
+static void
+clear_address (UbuntuGeoipProvider * provider)
+{
+	if (provider->priv->country_code == NULL &&
+			provider->priv->country_name == NULL &&
+			provider->priv->region_name == NULL &&
+			provider->priv->city == NULL &&
+			provider->priv->postal_code == NULL &&
+			provider->priv->timezone == NULL) {
+		/* We don't have any data to clear */
+		return;
+	}
+
+	if (provider->priv->country_code != NULL) {
+		g_free(provider->priv->country_code);
+		provider->priv->country_code = NULL;
+	}
+
+	if (provider->priv->country_name != NULL) {
+		g_free(provider->priv->country_name);
+		provider->priv->country_name = NULL;
+	}
+
+	if (provider->priv->region_name != NULL) {
+		g_free(provider->priv->region_name);
+		provider->priv->region_name = NULL;
+	}
+
+	if (provider->priv->city != NULL) {
+		g_free(provider->priv->city);
+		provider->priv->city = NULL;
+	}
+
+	if (provider->priv->postal_code != NULL) {
+		g_free(provider->priv->postal_code);
+		provider->priv->postal_code = NULL;
+	}
+
+	if (provider->priv->timezone != NULL) {
+		g_free(provider->priv->timezone);
+		provider->priv->timezone = NULL;
+	}
+
+	g_debug("Address Cleared");
+
+	GHashTable * address = geoclue_address_details_new();
+	GeoclueAccuracy * accuracy = geoclue_accuracy_new (GEOCLUE_ACCURACY_LEVEL_NONE, 0.0, 0.0);
+
+	gc_iface_address_emit_address_changed(GC_IFACE_ADDRESS(provider), 
+	                                        time(NULL),
+	                                        address,
+	                                        accuracy);
+
+	geoclue_accuracy_free(accuracy);
+	g_hash_table_unref(address);
+
+	return;
+}
+
 /* Callback for when the Network Manger state changes */
 static void
 nm_state_change (NMClient *client, const GParamSpec *pspec, gpointer user_data)
@@ -565,20 +753,25 @@ nm_state_change (NMClient *client, const GParamSpec *pspec, gpointer user_data)
 	g_return_if_fail(provider != NULL);
 
 	NMState state = nm_client_get_state (provider->priv->client);
-	gboolean connected = !(state == NM_STATE_CONNECTING || state == NM_STATE_DISCONNECTED);
 
-	if (provider->priv->connected != connected) {
+	if (provider->priv->network_state != state) {
 		/* Either way we want to refresh */
 		provider->priv->gotdata = FALSE;
-		if (connected) {
+		provider->priv->network_state = state;
+
+		clear_position(provider);
+		clear_address(provider);
+
+		if (state == NM_STATE_CONNECTED_GLOBAL) {
 			g_debug("Network connected");
 			gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_ACQUIRING);
-			provider->priv->connected = TRUE;
 			send_message(provider);
+		} else if (state == NM_STATE_CONNECTING) {
+			g_debug("Network connecting");
+			gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_ACQUIRING);
 		} else {
 			g_debug("Network lost");
 			gc_iface_geoclue_emit_status_changed(GC_IFACE_GEOCLUE(provider), GEOCLUE_STATUS_UNAVAILABLE);
-			provider->priv->connected = FALSE;
 		}
 	}
 
